@@ -1,7 +1,7 @@
 ;;;
 ;;; common-macros.scm - common macros
 ;;;
-;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2018  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,17 +38,58 @@
 ;; on other autoloaded files, since it is easy to create circular
 ;; dependency.
 
-(select-module gauche)
+(define-module gauche.common-macros
+  (export syntax-error syntax-errorf
+          push! pop! inc! dec! update!
+          check-arg get-optional get-keyword*
+          ^ ^_ ^a ^b ^c ^d ^e ^f ^g ^h ^i ^j ^k ^l ^m ^n ^o ^p ^q
+          ^r ^s ^t ^u ^v ^w ^x ^y ^z $
+          let1 if-let1 and-let1 rlet1
+          let/cc begin0 fluid-let
+          let-values let*-values
+          values-ref values->list
+          ecase
+          dotimes doilst doplist while until
+          guard unwind-protect
+          cond-list
+          assume assume-type))
+(select-module gauche.common-macros)
 
-;;; syntax-error
-;;; syntax-errorf
-;;;   Signals an error at compile time.
+;;; syntax-error msg arg ...
+;;; syntax-errorf fmtstr arg ...
+;;;   Signal an error at compile time.
+;;;   These are typically used as a result of expansion of syntax-rules
+;;;   macro; er-macro or legacy macro can directly call error/errorf so
+;;;   there's no point to use syntax-error.  Then, the 'original attribute
+;;;   of the form contains the macro input that caused syntax error.
+;;;   We extract that and throw a compound condition, so that the
+;;;   error message will include the macro input that directly caused
+;;;   this error.
+;;;   TODO: Move this to core after 0.9.6.
 
-(define-macro (syntax-error . args)
-  (apply error (map unwrap-syntax args)))
+(define-syntax syntax-error
+  (er-macro-transformer
+   (lambda (f r c)
+     (let ([args (map unwrap-syntax (cdr f))]
+           [original ((with-module gauche.internal pair-attribute-get)
+                      f 'original #f)])
+       (if original
+         (raise (make-compound-condition
+                 (apply make-error (car args) (cdr args))
+                 (make <compile-error-mixin> :expr original)))
+         (apply error args))))))
 
-(define-macro (syntax-errorf . args)
-  (apply errorf (map unwrap-syntax args)))
+(define-syntax syntax-errorf
+  (er-macro-transformer
+   (lambda (f r c)
+     (let ([args (map unwrap-syntax (cdr f))]
+           [original ((with-module gauche.internal pair-attribute-get)
+                      f 'original #f)])
+       (if original
+         (raise (make-compound-condition
+                 (make-error (apply format/ss (car args) (cdr args)))
+                 (make <compile-error-mixin> :expr original)))
+         (apply errorf args))))))
 
 ;;;-------------------------------------------------------------
 ;;; generalized set! family
@@ -248,6 +289,38 @@
                (lambda () ,@body)
                ,swap))))))
 
+;; srfi-11, r7rs
+(define-syntax let*-values
+  (syntax-rules ()
+    [(_ () . body)
+     (let () . body)]
+    [(_ ((formals init) . rest) . body)
+     (receive formals init
+       (let*-values rest . body))]))
+
+(define-syntax let-values
+  (er-macro-transformer
+   (^[f r c]
+     ;; we rename all the formals and rebind
+     (let* ([bindings (cadr f)]
+            [body (cddr f)]
+            [formals  (map car bindings)]
+            [inits    (map cadr bindings)]
+            [formals* (map (^f (map* (^x (gensym (x->string x)))
+                                     (^t (if (null? t) t (gensym (x->string t))))
+                                     f))
+                           formals)]
+            [rebinds (append-map
+                      (^[f f*]
+                        (map* list
+                              (^[a b] (if (null? a) '() (list (list a b))))
+                              f f*))
+                      formals formals*)])
+       (quasirename r
+         (let*-values ,(map list formals* inits)
+           (let ,rebinds
+             ,@body)))))))
+
 ;;;-------------------------------------------------------------
 ;;; applications
 
@@ -271,10 +344,17 @@
 ;;  ($ f a b $* g c d $*) => (lambda args (apply f a b (apply g c d args)))
 ;;  ($ f a b $ g c d $*) => (lambda args (f a b (apply g c d args)))
 
-(define-syntax $
+;; Kludge: We already have binding of '$' in gauche module, created for
+;; autoload.  Using (define-syntax $ ...) here makes a separate binding in
+;; gauche.common-macros, and that makes the literal comparison of '$'
+;; fail, since '$' in here refers to gauche.common-macros#$, while
+;; the macro use environment will refer to gauche#$.  We use very ugly
+;; hack here to workaround the issue; needs fundamental fix later.
+(define-syntax %$
   (syntax-rules ()
     [($ x . xs) (%$-split (x . xs) () ())]
     [($) (syntax-error "invalid $ form" ($))]))
+(set! $ %$) ; DON'T DO THIS.  Only works in the current version.
 
 ;; ($ a b $ c d $*) => ($* (c d) $ (a b))
 ;; ($ a b $ c d $* e f) => ((e f) $* (c d) $ (a b))
@@ -374,6 +454,30 @@
     [(_ . other)
      (syntax-error "malformed dolist" (dolist . other))]))
 
+(define-syntax doplist
+  (syntax-rules ()
+    [(_ ((k v) plis default) . body)
+     (do ([p plis (cddr p)])
+         [(cond [(null? p) #t]
+                [(null? (cdr p))
+                 (let ([k (car p)]
+                       [v default])
+                   . body)]
+                [else #f])]
+       (let ([k (car p)]
+             [v (cadr p)])
+         . body))]
+    [(_ ((k v) plis) . body)
+     (do ([p plis (cddr p)])
+         [(cond [(null? p) #t]
+                [(null? (cdr p)) (error "plist is not even:" plis)]
+                [else #f])]
+       (let ([k (car p)]
+             [v (cadr p)])
+         . body))]
+    [(_ . other)
+     (syntax-error "malformed doplist" (doplist . other))]))
+
 (define-syntax while
   (syntax-rules (=>)
     [(_ expr guard => var . body)
@@ -435,6 +539,8 @@
 ;;;-------------------------------------------------------------
 ;;; guard (srfi-34)
 
+(define %reraise (with-module gauche.internal %reraise))
+
 (define-syntax guard
   (syntax-rules ()
     [(guard (var . clauses) . body)
@@ -448,7 +554,8 @@
 (define-syntax %guard-rec
   (syntax-rules (else =>)
     [(%guard-rec var exc)
-     (raise exc)]
+     ;; exception handler can return to the caller
+     (%reraise)]
     [(%guard-rec var exc (else . exprs))
      (begin . exprs)]
     [(%guard-rec var exc (test => proc) . more)
@@ -476,24 +583,41 @@
 ;; deal with "ignore-errors" idiom, e.g. (guard (e [else #f]) body).  The exit
 ;; condition shouldn't be stopped in such a way.
 ;;
-;; TODO: Currently definition doesn't work when unwind-protect is used
+;; TODO: Current definition doesn't work when unwind-protect is used
 ;; within a thread that is terminated; thread termination isn't a condition
 ;; either.
 (define-syntax unwind-protect
   (syntax-rules ()
     [(unwind-protect body handler ...)
      (let ([x (exit-handler)]
-           [h (lambda () handler ...)])
+           [h (lambda () handler ...)]
+           [done #f])
        (with-error-handler
-           (lambda (e) (exit-handler x) (h) (raise e))
+           (lambda (e)
+             (exit-handler x)
+             (cond
+              [(condition-has-type? e <serious-condition>)
+               (unless done (set! done #t) (h))
+               ;; NB: We don't know E is thrown by r7rs#raise or
+               ;; r7rs#raise-continuable, but gauche#raise can handle both
+               ;; case.
+               (raise e)]
+              [else
+               ;; exception handler can return to the caller
+               (%reraise)]))
          (lambda ()
            (receive r
                (dynamic-wind
                  (lambda ()
-                   (exit-handler (lambda (code fmt args) (h) (x code fmt args))))
+                   (when done
+                     (error "Attempt to reenter obsoleted dynamic environment"))
+                   (exit-handler (lambda (code fmt args)
+                                   (set! done #t)
+                                   (h)
+                                   (x code fmt args))))
                  (lambda () body)
-                 (lambda ()
-                   (exit-handler x)))
+                 (lambda () (exit-handler x)))
+             (set! done #t)
              (h)
              (apply values r)))
          :rewind-before #t))]
@@ -538,13 +662,24 @@
        (if tmp (cons (begin . expr) r) r))]
     ))
 
-;;; ------------------------------------------------------------
-;;; er-macro-transformer
+;;;----------------------------------------------------------------
+;;; assume (srfi-145) and co.
 ;;;
 
-;; This will become compiled into core after 0.9.5 release; for 0.9.4,
-;; we can't compile-in define-syntaxed macros, so we keep this autoload.
-(define-syntax er-macro-transformer
-  (primitive-macro-transformer
-   (^[form def-env use-env]
-     ((with-module gauche.internal %expand-er-transformer) form use-env))))
+;; We might add run-time optimization switch to expand assume to nothing.
+(define-syntax assume
+  (syntax-rules ()
+    [(_ expr)
+     (unless expr
+       (error (format "Invalid assumption: ~s" 'expr)))]
+    [(_ expr message ...)
+     (unless expr
+       (error (format "Invalid assumption: ~s:" 'expr) message ...))]))
+
+(define-syntax assume-type
+  (syntax-rules ()
+    [(_ expr type)
+     (let1 v expr
+       (unless (is-a? v type)
+         (type-error 'expr type v)))]
+    ))
