@@ -1,7 +1,7 @@
 ;;;
 ;;; mime-port.scm - submodule to read from mime part body
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2018  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -36,7 +36,7 @@
 (define-module rfc.mime-port
   (use gauche.uvector)
   (use gauche.vport)
-  (use util.queue)
+  (use data.queue)
   (use util.match)
   (export make-mime-port))
 (select-module rfc.mime-port)
@@ -54,7 +54,7 @@
 ;; either EOF or MIME boundary.  Basically it runs a DFA.
 (define (make-mime-port boundary srcport)
   (define q (make-queue))
-  (define --boundary (string->u8vector #`"--,boundary"))
+  (define --boundary (string->u8vector #"--~boundary"))
 
   (define port (make <mime-port>))
 
